#!/usr/bin/env python3

import importlib.metadata
import os
from enum import Enum
from pathlib import Path
from typing import Callable, Iterable, Optional

import requests
import typer

import mandown.sources
from mandown import mandown
from mandown.converter import Converter
from mandown.sources.base_source import BaseSource

app = typer.Typer()


class ConvertFormats(str, Enum):
    CBZ = "cbz"
    EPUB = "epub"
    MOBI = "mobi"
    PDF = "pdf"
    FOLDER = "folder"


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"mandown {importlib.metadata.version('mandown')}")
        raise typer.Exit()


def supported_sites_callback(value: bool) -> None:
    if value:
        for source in mandown.sources.get_all_classes():
            typer.echo(f"{source.name}: {', '.join(source.domains)}")
        raise typer.Exit()


# pylint: disable=unused-argument
@app.command()
def download(
    url: str,
    dest: str = typer.Argument(
        os.getcwd(), help="The destination folder to download to."
    ),
    convert: ConvertFormats = typer.Option(
        "folder", "--convert", "-c", help="The format to download the comic as"
    ),
    from_folder: Path = typer.Option(
        None,
        "--from",
        help="The folder to assume the comic exists in to convert without downloading",
    ),
    start: Optional[int] = typer.Option(
        None,
        "--start",
        "-s",
        help="The first chapter to download [default: first found]",
    ),
    end: Optional[int] = typer.Option(
        None, "--end", "-e", help="The last chapter to download [default: last found]"
    ),
    maxthreads: int = typer.Option(
        4,
        "--threads",
        "-t",
        help="The maximum number of images to download in parallel",
    ),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        is_eager=True,
        help="Display the current version of mandown",
    ),
    supported_sites: Optional[bool] = typer.Option(
        None,
        "--supported-sites",
        callback=supported_sites_callback,
        is_eager=True,
        help="Output a list of domains supported by mandown",
    ),
) -> None:
    """
    Download from a URL chapters start_chapter to end_chapter.
    Defaults to the first chapter and last chapter, respectively
    in the working directory.
    """
    if not os.path.isdir(dest):
        raise ValueError(f"{dest} is not a valid folder path.")

    typer.echo(f"Searching sources for {url}")
    try:
        source: BaseSource = mandown.query(url, populate=True)
        typer.secho(f'Found item from source "{source.name}"', fg=typer.colors.GREEN)
        typer.secho(source)
    except ValueError as err:
        typer.secho("Could not match URL with available sources.", fg=typer.colors.RED)
        raise typer.Exit(1) from err

    # starting to think that immutability is much better than whatever
    # the heck is going on here
    target_path = os.path.join(dest, source.metadata.title)
    if not os.path.isdir(target_path):
        os.mkdir(target_path)

    # if they are undefined
    start_chapter = start or 1
    end_chapter = end or len(source.chapters)

    # zero-index
    start_chapter -= 1

    chapter_range = []
    if (
        from_folder is None
    ):  # yikes these should be split into their own functions sometime
        # get cover art
        with open(Path(target_path) / "cover.jpg", "w+b") as file:
            file.write(requests.get(source.metadata.cover_art).content)

        chapter_range = source.chapters[start_chapter:end_chapter]
        typer.echo("Downloading...")
        for i, chapter in enumerate(chapter_range):
            with typer.progressbar(
                mandown.download_chapter_progress(chapter, target_path, maxthreads),
                length=len(chapter.images),
                label=f"{chapter.title} ({i+1}/{len(chapter_range)})",
            ) as progress:
                for _ in progress:
                    pass
        typer.secho(
            f"Successfully downloaded {len(chapter_range)} chapters.",
            fg=typer.colors.GREEN,
        )
    else:
        if convert.value == ConvertFormats.FOLDER:
            typer.secho(
                "A convert format was not specified but is required for conversion.",
                fg=typer.colors.RED,
            )
            raise typer.Exit(1)
        if start or end:
            typer.secho(
                "WARNING: --start and --end have no effect when --from is used",
                fg=typer.colors.BRIGHT_RED,
            )
        target_path = str(from_folder)

    if convert.value != ConvertFormats.FOLDER:
        typer.echo(f"Converting to {convert.value}...")

        formatted_chapters = [(c.title, c.title_sanitised) for c in chapter_range]
        converter = Converter(
            target_path,
            source.metadata,
            chapter_list=None if from_folder else formatted_chapters,
        )

        convert_func: Callable[[str], Iterable] = lambda i: None
        match convert.value:
            case ConvertFormats.EPUB:
                convert_func = converter.to_epub_progress
            case ConvertFormats.CBZ:
                convert_func = converter.to_cbz_progress
            case ConvertFormats.MOBI:
                raise ValueError("MOBI conversion is not yet supported.")
            case ConvertFormats.PDF:
                raise ValueError("PDF conversion is not yet supported.")

        with typer.progressbar(
            length=converter.max_operations[convert.value], label="Converting"
        ) as progress:
            for i in convert_func(dest):
                progress.update(1, i)

        dest_file = dest / Path(source.metadata.title).with_suffix(f".{convert.value}")
        typer.secho(f"Successfully converted to {dest_file}", fg=typer.colors.GREEN)


def main() -> None:
    app()


if __name__ == "__main__":
    main()



# pylint: disable=invalid-name
import os
from pathlib import Path
from typing import Iterable

from natsort import natsorted

from mandown import iohandler, sources
from mandown.sources.base_source import BaseSource, Chapter


def query(url: str, populate: bool = True, populate_sort: bool = True) -> BaseSource:
    """
    Return the source file for a URL.
    """
    source = sources.get_class_for(url)(url)
    if populate:
        # these statements are to trigger their getters to
        # fetch the data
        if source.metadata:
            pass
        if source.chapters:
            if populate_sort:
                titles = list(map(lambda c: c.title, source.chapters))
                if titles != natsorted(titles):
                    padding = f"0{len(str(len(source.chapters)))}"
                    for i, c in enumerate(source.chapters):
                        c.title = f"{i+1:{padding}}. {c.title}"
    return source


def download_chapter_progress(
    chapter: Chapter, dest_folder: str, maxthreads: int = 1, only_needed: bool = True
) -> Iterable[None]:
    """
    Download the images of a chapter to a destination folder.
    Returns a generator that increments whenever an item has finished
    downloading.
    Raises ValueError if the folder does not exist.
    """
    if not chapter.images:
        raise ValueError("No images available to download")

    if not os.path.isdir(dest_folder):
        raise ValueError(f"Folder path {dest_folder} does not exist")

    download_folder = os.path.join(dest_folder, chapter.title)
    if not os.path.isdir(download_folder):
        os.mkdir(download_folder)

    padding = len(str(len(chapter.images)))
    skip_images: set[int] = set()
    if only_needed:
        for f in os.listdir(download_folder):
            name = Path(f).stem
            if name == name.rjust(padding, "0"):
                try:
                    skip_images.add(int(name))
                except ValueError:
                    # expected if it's not an image
                    pass

    if len(skip_images) != len(chapter.images):
        # zip will crash if fed an empty array
        processed_chapter_images, filestems = zip(
            *(
                (link, str(i).rjust(padding, "0"))
                for i, link in enumerate(chapter.images, start=1)
                if i not in skip_images
            )
        )

        yield from iohandler.download(
            processed_chapter_images,
            os.path.join(dest_folder, download_folder),
            chapter.headers,
            maxthreads,
            filestems,
        )


def download_chapter(
    chapter: Chapter, dest_folder: str, maxthreads: int = 1, only_needed: bool = True
) -> None:
    """
    Download the images of a chapter to a destination folder.
    Raises ValueError if the folder does not exist.
    """
    for _ in download_chapter_progress(chapter, dest_folder, maxthreads, only_needed):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mandown', 'mandown.sources']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.10.0,<5.0.0',
 'feedparser>=6.0.8,<7.0.0',
 'lxml>=4.7.1,<5.0.0',
 'natsort>=8.0.2,<9.0.0',
 'requests>=2.27.0,<3.0.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['mandown = mandown.cli:main']}

setup_kwargs = {
    'name': 'mandown',
    'version': '0.6.1',
    'description': 'Command line application and library to download and convert comics from various sources',
    'long_description': '# mandown\n\n<a href="https://www.codefactor.io/repository/github/potatoeggy/mandown"><img src="https://www.codefactor.io/repository/github/potatoeggy/mandown/badge" alt="CodeFactor" /></a>\n<a href="https://pypi.org/project/mandown"><img src="https://img.shields.io/pypi/v/mandown" /></a>\n<a href="https://aur.archlinux.org/packages/mandown-git"><img src="https://img.shields.io/aur/version/mandown-git" /></a>\n<a href="https://github.com/potatoeggy/mandown/releases/latest"><img src="https://img.shields.io/github/v/release/potatoeggy/mandown?display_name=tag" /></a>\n<a href="/LICENSE"><img src="https://img.shields.io/github/license/potatoeggy/mandown" /></a>\n\nComic downloader and converter to CBZ and EPUB as a Python library and command line application.\n\n## Supported sites\n\n- https://mangasee123.com\n- https://manganato.com\n- https://webtoons.com\n- https://mangadex.org\n- https://mangakakalot.com\n\n## Installation\n\nInstall the package from PyPI:\n\n```\npip3 install mandown\n```\n\nOr, to build from source:\n\nMandown depends on [poetry](https://github.com/python-poetry/poetry) for building.\n\n```\ngit clone https://github.com/potatoeggy/mandown.git\npoetry install\npoetry build\npip3 install dist/mandown*.whl\n```\n\n## Usage\n\n```\nmandown <URL>\n```\n\nTo convert the download contents to CBZ/EPUB, append the `--convert` option.\n\n```\nmandown <URL> --convert EPUB\n```\n\nTo only\n\nRun `mandown --help` for more info.\n\n## Library usage\n\n```python\nimport os\nfrom mandown import mandown\n\nmanga = mandown.query(url_to_manga)\nprint(manga.metadata, manga.chapters)\nfor c in manga.chapters:\n    mandown.download_chapter(c, dest_folder=os.getcwd(), maxthreads=4)\n```\n',
    'author': 'Daniel Chen',
    'author_email': 'danielchen04@hotmail.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/potatoeggy/mandown',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Converts a folder of images to EPUB/PDF/CBZ/MOBI
"""
import datetime
import os
import re
import shutil
import tempfile
import textwrap
import unicodedata
from pathlib import Path
from typing import Iterator, Optional
import zipfile

from natsort import natsorted

from mandown.sources.base_source import MangaMetadata

ACCEPTED_IMAGE_EXTENSIONS = {
    ".gif": "gif",
    ".png": "png",
    ".jpg": "jpeg",
    ".jpeg": "jpeg",
    ".jpe": "jpeg",
}


class Converter:
    def __init__(
        self,
        folder_path: str,
        metadata: Optional[MangaMetadata] = None,
        chapter_list: Optional[list[tuple[str, str]]] = None,
    ) -> None:
        """
        Get the file path of a comic and identify its metadata.
        The chapter list is a list of tuples in the format
        (`chapter_path`, `chapter_title`), where `chapter_path`
        is the path of the chapter relative to `folder_path`.

        If the chapter list or metadata are not provided, they will
        be guessed from the directory structure.
        """
        if not Path(folder_path).is_dir():
            raise ValueError(f"{folder_path} is not a directory.")

        self.folder_path = Path(os.path.realpath(folder_path))
        self.metadata = metadata or MangaMetadata(
            title=os.path.basename(self.folder_path),
            authors=[],
            url="",
            genres=[],
            description="",
            cover_art="",
        )

        # self.chapters is found either by traversing the path (natsort) or by being fed
        working_chapters = chapter_list or []
        if not working_chapters:
            for item in natsorted(os.listdir(self.folder_path)):
                item = str(item)  # get rid of type complaints
                if (Path(self.folder_path) / item).is_dir():
                    working_chapters.append((item, item))

        padding = f"0{len(str(len(working_chapters)))}"

        self.chapters: list[tuple[str, str, str, list[str]]] = [
            (
                title,
                sanitised,
                f"{i+1:{padding}}-{self.slugify(sanitised)}",
                list(
                    filter(
                        lambda s: any(
                            str(s).lower().endswith(i)
                            for i in ACCEPTED_IMAGE_EXTENSIONS
                        ),
                        natsorted(os.listdir(self.folder_path / sanitised)),  # type: ignore
                    )
                ),
            )
            for i, (title, sanitised) in enumerate(working_chapters)
        ]

        # useful for progress bars
        self.max_operations: dict[str, int] = {
            "epub": len(self.chapters) * 2,
            "cbz": len(self.chapters),
        }

    def to_cbz_progress(self, dest_folder: str) -> Iterator:
        dest_file = Path(dest_folder) / f"{self.metadata.title}.cbz"
        with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as file:  # type: ignore
            for dirpath, _, filenames in os.walk(self.folder_path):  # from kcc
                yield "Compressing"
                for name in filenames:
                    if (Path(dirpath) / name).is_file():
                        file.write(
                            Path(dirpath) / name,  # type: ignore
                            Path(dirpath.lstrip(str(self.folder_path))) / name,
                            zipfile.ZIP_DEFLATED,
                        )

    def to_cbz(self, dest_folder: str) -> None:
        for _ in self.to_cbz_progress(dest_folder):
            pass

    def to_epub_progress(self, dest_folder: str) -> Iterator:
        # taken from KCC generated epub file structure
        with tempfile.TemporaryDirectory() as temp:
            root = Path(temp)
            EpubGenerator.create_skeleton(root)

            oebps = root / "OEBPS"
            with open(oebps / "Text" / "style.css", "w", encoding="utf-8") as file:
                file.write(EpubGenerator.style_css)

            with open(oebps / "content.opf", "w", encoding="utf-8") as file:
                file.write(
                    EpubGenerator.generate_content_opf(self.metadata, self.chapters)
                )

            with open(oebps / "toc.ncx", "w", encoding="utf-8") as file:
                file.write(EpubGenerator.generate_toc_ncx(self.metadata, self.chapters))

            with open(oebps / "nav.xhtml", "w", encoding="utf-8") as file:
                file.write(
                    EpubGenerator.generate_nav_xhtml(self.metadata, self.chapters)
                )

            for _, path, slug, images in self.chapters:
                yield "Building"
                os.mkdir(oebps / "Text" / slug)
                os.mkdir(oebps / "Images" / slug)
                padding = f"0{len(str(len(images)))}"

                for index, i in enumerate(images, start=1):
                    with open(
                        oebps / "Text" / slug / f"{index:{padding}}.xhtml",
                        "w",
                        encoding="utf-8",
                    ) as file:
                        file.write(
                            EpubGenerator.generate_image_html(
                                slug, f"{index:{padding}}{Path(i).suffix}"
                            )
                        )
                    shutil.copyfile(
                        self.folder_path / path / i,
                        oebps / "Images" / slug / f"{index:{padding}}{Path(i).suffix}",
                    )

            if (self.folder_path / "cover.jpg").is_file():
                shutil.copyfile(
                    self.folder_path / "cover.jpg", oebps / "Images" / "cover.jpg"
                )

            # compress and move epub
            dest_file = Path(dest_folder) / f"{self.metadata.title}.epub"

            with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as zipout:
                zipout.writestr("mimetype", "application/epub+zip", zipfile.ZIP_STORED)
                for dirpath, _, filenames in os.walk(root):  # from kcc
                    yield "Compressing"
                    for name in filenames:
                        if (Path(dirpath) / name).is_file():
                            zipout.write(
                                Path(dirpath) / name,
                                Path(dirpath.lstrip(str(root))) / name,
                                zipfile.ZIP_DEFLATED,
                            )

    def to_epub(self, dest_folder: str) -> None:
        for _ in self.to_epub_progress(dest_folder):
            pass

    def to_pdf(self) -> None:
        pass

    def to_mobi(self) -> None:
        pass

    @staticmethod
    def slugify(string: str) -> str:
        """
        Remove all potentially invalid characters for XML
        """
        slug = unicodedata.normalize("NFKD", string)
        slug = slug.lower()
        slug = re.sub(r"[^a-z0-9]+", "-", slug).strip("-")
        slug = re.sub(r"[-]+", "-", slug)
        return slug


# pylint: disable=line-too-long
class EpubGenerator:
    style_css = textwrap.dedent(
        """\
        @page {
        margin: 0;
        }
        body {
        display: block;
        margin: 0;
        padding: 0;
        }\
        """
    )

    @staticmethod
    def create_skeleton(root: Path) -> None:
        os.mkdir(root / "META-INF")
        with open(root / "META-INF" / "container.xml", "w", encoding="utf-8") as file:
            file.write(
                textwrap.dedent(
                    """\
                    <?xml version="1.0"?>
                    <container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">
                    <rootfiles>
                    <rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/>
                    </rootfiles>
                    </container>\
                    """
                )
            )

        os.mkdir(root / "OEBPS")
        os.mkdir(root / "OEBPS" / "Images")
        os.mkdir(root / "OEBPS" / "Text")

    @staticmethod
    def generate_image_html(
        chapter_slug: str, image_path: str, width: int = 1005, height: int = 1430
    ) -> str:
        image_base = Path(image_path).with_suffix("")
        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="UTF-8"?>
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">
            <head>
            <title>{image_base}</title>
            <link href="../style.css" type="text/css" rel="stylesheet"/>
            <meta name="viewport" content="width={width}, height={height}"/>
            </head>
            <body style="">
            <div style="text-align:center;top:0.0%;">
            <img width="{width}" height="{height}" src="../../Images/{chapter_slug}/{image_path}"/>
            </div>
            </body>
            </html>\
            """
        )

    @staticmethod
    def generate_toc_ncx(
        metadata: MangaMetadata, chapters: list[tuple[str, str, str, list[str]]]
    ) -> str:
        nav_map: list[str] = []
        for title, _, slug, images in chapters:
            padding = f"0{(len(str(len(images))))}"
            nav_map.append(
                f'<navPoint id="Text_{slug}"><navLabel><text>{title}</text></navLabel><content src="Text/{slug}/{1:{padding}}.xhtml"/></navPoint>'
            )

        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="UTF-8"?>
            <ncx version="2005-1" xml:lang="en-US" xmlns="http://www.daisy.org/z3986/2005/ncx/">
            <head>
            <meta name="dtb:uid" content="urn:uuid:bafec70e-0286-49d3-b1e2-9e9f297c8cfe"/>
            <meta name="dtb:depth" content="1"/>
            <meta name="dtb:totalPageCount" content="0"/>
            <meta name="dtb:maxPageNumber" content="0"/>
            <meta name="generated" content="true"/>
            </head>
            <docTitle><text>{metadata.title}</text></docTitle>
            <navMap>
            {(chr(10) + " "*12).join(nav_map)}
            </navMap>
            </ncx>\
            """
        )

    @staticmethod
    def generate_content_opf(
        metadata: MangaMetadata, chapters: list[tuple[str, str, str, list[str]]]
    ) -> str:
        """
        `chapters` must be a list containing a tuple of elements in this order:
          - `title`, the title of the chapter
          - `sanitised_title`, the relative path to the chapter folder
          - `slug`, a safe-for-html version of the title
          - `images`, a list of paths to each image
        """

        time_now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()

        item_ids: list[str] = []
        item_refs: list[str] = []
        for *_, slug, images in chapters:
            padding = f"0{len(str(len(images)))}"
            for index, i in enumerate(images, start=1):
                marker = f"{index:{padding}}"
                item_ids.append(
                    f'<item id="page_Images_{slug}_{marker}" href="Text/{slug}/{marker}.xhtml" media-type="application/xhtml+xml" />'
                )
                item_ids.append(
                    f'<item id="img_Images_{slug}_{marker}" href="Images/{slug}/{marker}{Path(i).suffix}" media-type="image/{ACCEPTED_IMAGE_EXTENSIONS[Path(i).suffix]}" />'
                )

                item_refs.append(f'<itemref idref="page_Images_{slug}_{marker}" />')

        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="UTF-8"?>
            <package version="3.0" unique-identifier="BookID" xmlns="http://www.idpf.org/2007/opf">
            <metadata xmlns:opf="http://www.idpf.org/2007/opf" xmlns:dc="http://purl.org/dc/elements/1.1/">
            <dc:title>{metadata.title}</dc:title>
            <dc:language>en-US</dc:language>
            <dc:identifier id="BookID">urn:uuid:bafec70e-0286-49d3-b1e2-9e9f297c8cfe</dc:identifier>
            <dc:contributor id="contributor">mandown</dc:contributor>
            {(chr(10) + " "*12).join(
                f'<dc:creator id="id-{i}">{name}</dc:creator>'
                for i, name in enumerate(metadata.authors)
            )}
            <dc:description>{
                (chr(10) + " "*12).join(
                    metadata.description.split(chr(10))
                )
            }</dc:description>
            {
                (chr(10) + " "*12).join(
                    f"<dc:subject>{genre}</dc:subject>"
                    for genre in metadata.genres
                )
            }
            <meta property="dcterms:modified">{time_now}Z</meta>
            <meta name="cover" content="cover"/>
            <meta property="rendition:orientation">portrait</meta>
            <meta property="rendition:spread">both</meta>
            <meta property="rendition:layout">pre-paginated</meta>
            </metadata>
            <manifest>
            <item id="ncx" href="toc.ncx" media-type="application/x-dtbncx+xml"/>
            <item id="nav" href="nav.xhtml" properties="nav" media-type="application/xhtml+xml"/>
            <item id="css" href="Text/style.css" media-type="text/css"/>
            <item id="cover" href="Images/cover.jpg" media-type="image/jpeg" properties="cover-image"/>
            {(chr(10) + " "*12).join(item_ids)}
            </manifest>
            <spine page-progression-direction="ltr" toc="ncx">
            {(chr(10) + " "*12).join(item_refs)}
            </spine>
            </package>\
            """
        )

    @staticmethod
    def generate_nav_xhtml(
        metadata: MangaMetadata, chapters: list[tuple[str, str, str, list[str]]]
    ) -> str:
        lis: list[str] = []
        for title, _, slug, images in chapters:
            padding = f"0{len(str(len(images)))}"
            lis.append(
                f'<li><a href="Text/{slug}/{1:{padding}}.xhtml">{title}</a></li>'
            )

        return textwrap.dedent(
            f"""\
            <?xml version="1.0" encoding="utf-8"?>
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">
            <head>
            <title>{metadata.title}</title>
            <meta charset="utf-8"/>
            </head>
            <body>
            <nav xmlns:epub="http://www.idpf.org/2007/ops" epub:type="toc" id="toc">
            <ol>
            {(chr(10) + " "*12).join(lis)}
            </ol>
            </nav>
            <nav epub:type="page-list">
            <ol>
            {(chr(10) + " "*12).join(lis)}
            </ol>
            </nav>
            </body>
            </html>\
            """
        )
